import sys
from typing import Union

from PySide6 import QtWidgets, QtCore
from PySide6.QtWidgets import QApplication, QWidget, QDialog, QLabel, QFrame

from sohstationviewer.conf import constants


def add_separation_line(layout):
    """
    Add a line for separation to the given layout.
    :param layout: QLayout - the layout that contains the line
    """
    label = QLabel()
    label.setFrameStyle(QFrame.Shape.HLine | QFrame.Shadow.Sunken)
    label.setLineWidth(1)
    layout.addWidget(label)


class AboutDialog(QDialog):
    """
     Dialog to show information of the software.

     About Dialog is always opened and will be raised up when the about menu
     action is triggered.
    """
    def __init__(self, parent: Union[QWidget, QApplication]):
        """
        :param parent: the parent widget
        """
        super(AboutDialog, self).__init__(parent)
        # set block interaction with other windows
        self.setWindowModality(QtCore.Qt.WindowModality.ApplicationModal)
        self.software_name_label = QLabel("SOHViewer")
        self.software_name_label.setStyleSheet(
            "QLabel {font-size: 18pt; font-style: bold; color: darkblue}"
        )
        description = ("Visualize State-of-Health packets from data in "
                       "mseed or reftek formats recorded\n"
                       "by different types of data loggers.")
        self.description_label = QLabel(description)

        version = f"Version {constants.SOFTWARE_VERSION}"
        self.version_label = QLabel(version)

        built_time = f"Built on {constants.BUILD_TIME}"
        self.built_time_label = QLabel(built_time)

        copyright = u"Copyright \u00A9 2024 EarthScope Consortium"
        self.copyright_label = QLabel(copyright)

        self.ok_button = QtWidgets.QPushButton('OK', self)
        self.setup_ui()
        self.connect_signals()

    def setup_ui(self):
        self.setWindowTitle("About SOHViewer")
        main_layout = QtWidgets.QVBoxLayout()
        self.setLayout(main_layout)

        main_layout.addWidget(self.software_name_label)
        main_layout.addWidget(self.description_label)
        add_separation_line(main_layout)

        main_layout.addWidget(self.version_label)
        main_layout.addWidget(self.built_time_label)
        main_layout.addWidget(self.copyright_label)

        button_layout = QtWidgets.QHBoxLayout()
        main_layout.addLayout(button_layout)
        button_layout.addStretch()
        button_layout.addWidget(self.ok_button)

    def connect_signals(self) -> None:
        self.ok_button.clicked.connect(self.close)


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    test = AboutDialog(None)
    test.exec()
    sys.exit(app.exec())
